/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/

package casa.dodwan.napi.ws;


import casa.dodwan.napi.DodwanNapiEndpoint;
import casa.util.pdu.BsonPduSerializer;
import casa.dodwan.napi.DodwanNapiServer;
import casa.util.pdu.JsonPduSerializer;
import casa.util.pdu.PduSerializer;
import casa.dodwan.run.Dodwan;
import casa.dodwan.run.dodwand;
import casa.dodwan.util.ServiceEnvironment;
import casa.util.Logger;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.java_websocket.WebSocket;
import org.java_websocket.handshake.ClientHandshake;
import org.java_websocket.server.WebSocketServer;



/**
 * DoDWAN plugin that provides a lightweight websocket server for the DoDWAN Network API
 * Should be a singleton.
 * 
 * The server can handle multiple connections (in DodwanEndpoints).
 * A DodwanEndpoint is named after the path of the URI that is used to open the websocket connection.
 * The server offers a static method to retrieve a DodwanEndpoint according to its name.
 */

public class DodwanWsServer implements DodwanNapiServer {
   
    
    // Static... pour faire simple.
    static public Dodwan dodwan = null;
    static public PduSerializer pduSerializer = null;
    static public Map<WebSocket, DodwanWsEndpoint> endpointsBySocket; 
    static public Map<String, DodwanWsEndpoint> endpointsByName; 
   
    public String host = null;

    private dodwand dodwand;
    private ServiceEnvironment dodwanWsServerEnvironment;
    
    // Port number
    public int port;
    
    
    private WsServer server;

     // Connection timeout (in seconds)
     //  -1 for no timeout
    public int connectionTimeout;
      
    
    
    //-----------------------------------------------------------------------------
     
    public DodwanWsServer(dodwand dodwand) {
       this.dodwand = dodwand;
       this.dodwan = dodwand.dodwan;    
       
       dodwanWsServerEnvironment = new ServiceEnvironment("dodwan_napi_ws");
       this.host = dodwanWsServerEnvironment.properties.getString("dodwan_napi_ws.host", "localhost");
       this.port = dodwanWsServerEnvironment.properties.getInteger("dodwan_napi_ws.port", 8025);
       this.connectionTimeout = dodwanWsServerEnvironment.properties.getInteger("dodwan_napi_ws.conn_timeout", 50);
               
       String serializingMethod = dodwanWsServerEnvironment.properties.getString("dodwan_napi_ws.serial_method", "json");
       switch (serializingMethod) {
           case "json":
               this.pduSerializer = new JsonPduSerializer();
               break;
            case "bson":
               this.pduSerializer = new BsonPduSerializer();
               break;  
            default:
                System.err.println("DodwanWsServer: unknown serializing method "+ serializingMethod
                                   + ". Falling back to bson");
                this.pduSerializer = new BsonPduSerializer();
       }
              
      
           
      InetSocketAddress address = new InetSocketAddress(this.host, this.port);
      
      this.server= new WsServer(address);
      
      
 
    }

    /**
     * Gives the list of endpoints handled by the server 
     * @return the list of DoDWAN WS endpoints 
    
    static public List<DodwanNapiEndpoint> getEndpoints() {
        return new ArrayList(DodwanWsServer.endpointsBySocket.values());
    }
     */
    
    
    /**
     * Returns the endpoint according to its name 
     * @return the endpoint or null if it does not exist
     */
    static public DodwanNapiEndpoint getEndpoint(String name) {
        return DodwanWsServer.endpointsByName.get(name);
    }
    
     
    
    
    
    //-----------------------------------------------------------------------------
    /**
     * Gives the number of the real port the server listens to. 
     * It is the asked port or the port chosen by the system.
     * @return the port number 
     */
    public int getPort() {
        return this.port;
    }
    
      
    //-----------------------------------------------------------------------------
    public void start() {
        this.server.start();
        
        if (this.port == 0) {
          System.err.println("DodwanWsServer waiting for dynamic port choice");
          int rounds = 0;
          while ((this.port = this.server.getPort()) <= 0) {
             try {
                 rounds ++;
                 Thread.sleep(5);
                 port = DodwanWsServer.this.server.getPort();
             } catch (InterruptedException ex) {
                 
             }
          } 
          System.err.println("DodwanWsServer waited for " + rounds + " rounds of 5ms. Got port "+this.port);
          dodwanWsServerEnvironment.properties.setInteger("dodwan_napi_ws.port", this.port);
       }
       System.out.println("DodwanWsServer is listening on port "+this.port);
       dodwand.writePortFile("dodwan_napi_ws.port", this.port);
      
       Logger.log("napi-ws","Starting TooTallNate websocket server. Listening on port "+this.port);
       System.err.println("TooTallNate websocket server started (port="+DodwanWsServer.this.port +
                    "  connection timeout="+DodwanWsServer.this.connectionTimeout+"s)");
    }
      
    
    
    //-----------------------------------------------------------------------------
    public void stop() {
        Logger.log("napi-ws","Stopping TooTallNate websocket server.");
        try {
            this.server.stop();
        } catch (Exception ex) {
             System.err.println("Error when stopping TooTallNate websocket server");
             ex.printStackTrace(System.err);
        }
    }
    
    
    
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    public class WsServer extends WebSocketServer {
  

        public WsServer(InetSocketAddress addr) {
            super(addr);
            DodwanWsServer.endpointsBySocket = new HashMap<WebSocket, DodwanWsEndpoint>();
            DodwanWsServer.endpointsByName = new HashMap<String, DodwanWsEndpoint>();
        }
        
                

        @Override
        public void onOpen(WebSocket conn, ClientHandshake handshake) {
            System.err.println("Opening websocket connection with client at " + conn.getRemoteSocketAddress().getAddress().getHostAddress());
            
            URI uri = null;
            String path=null;
            String query = null;
            String endpointName = "";
            try {
                uri = new URI(handshake.getResourceDescriptor());
                path = uri.getPath();
                query = uri.getQuery();
                System.err.println("URI Path:" + path);
                //System.err.println("URI Query:" + query);
            } catch (NullPointerException npe) {
                System.err.println("Null URI");
            }
            catch (URISyntaxException ex) {
                 System.err.println("Malformed URI: "+uri);
            }
            if (path!= null) {
                endpointName = path.substring(1);
            }
                        
            DodwanWsEndpoint endpoint = new DodwanWsEndpoint(conn, endpointName);
            DodwanWsServer.endpointsBySocket.put(conn, endpoint);
             System.err.println("Regsitering DoDWAN enpoint "+ endpointName);
            DodwanWsServer.endpointsByName.put(endpointName, endpoint);
            
        }

        @Override
        public void onClose(WebSocket conn, int code, String reason, boolean remote) {
            //if (conn != null) System.err.println("Closing websocket connection with client at " + conn.getRemoteSocketAddress().getAddress().getHostAddress());
            DodwanWsServer.endpointsBySocket.remove(conn);
        }

        @Override
        public void onMessage(WebSocket conn, String message) {
            if (conn != null && conn.getRemoteSocketAddress() != null)
                System.err.println("Strange: I received a string message from " + conn.getRemoteSocketAddress().getAddress().getHostAddress());
        }

        @Override
        public void onMessage(WebSocket conn, ByteBuffer message) {
            DodwanWsServer.endpointsBySocket.get(conn).rcvMsg(message);
        }

        @Override
        public void onError(WebSocket conn, Exception ex) {
            if (conn == null || conn.getRemoteSocketAddress() == null) {
                System.err.println("Websocket Error");
            } else {
                System.err.println("Websocket Error on connection with " + conn.getRemoteSocketAddress().getAddress().getHostAddress());
            }
            ex.printStackTrace(System.err);
        }

        @Override
        public void onStart() {
          
            setConnectionLostTimeout(DodwanWsServer.this.connectionTimeout*1000);
        }

                

    
    
    }
    
    
    

}